/*
homealone.c v0.91 Copyright (c) Julian Assange 1995, All rights reserved.

This software maybe used freely by registered non-profit organizations
and non-profit educational institutions. All other users and
organizations, included government bodies and organizations shall be
defined for the purposes of this document as commercial users.

Commercial users may not use this software without an appropriate
agreement or license with the Author. Please see the documentation on
LICENSING contained elsewhere with this software.

This software maybe distributed only freely, in full and without
modification.  It may not be bundled with any sort of hardware or
software if a fee is charged for that hardware or software directly or
indirectly for that hardware or software.  If you would like to include
this software in such a distribution then please contact the author to
negotiate reasonable (possibly free) terms.

The author shall not under any circumstances accept any liability for
this software, for its use, misuse, or any failings it may or may not
have. Your on your own.

The author reserves the right to alter the aformentioned conditions from
time to time as he sees appropriate. The author's most recent copyright
notice and conditions for this software always supersede any issued
previously.

Use and or distribution of this software implies acceptance of the above.
*/

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <pwd.h>
#include <ctype.h>
#include <time.h>
#include <string.h>
#include <dirent.h>
#include <fcntl.h>
#include <netdb.h>

#include "VERSION.h"
#include "SERIAL.h"

char *re_comp(char *);
int re_exec(char *);
extern char *optarg;

#ifndef S_IFMT
#  define S_IFMT  00170000
#endif
#ifndef S_IFREG
#  define S_IFREG  0100000
#endif
#ifndef S_ISREG
#  define S_ISREG(m)      (((m) & S_IFMT) == S_IFREG)
#endif
#ifndef S_IFDIR
#  define S_IFDIR 0040000
#endif
#ifndef S_ISDIR
#  define S_ISDIR(m)      (((m) & S_IFMT) == S_IFDIR)
#endif

#define LEFT 0
#define CENTER 1
#define RIGHT 2

char *align[]={"left", "center", "right"};

struct schain
{
	struct schain *next;
	char *str;
};

char *months[]={"Jan","Feb","Mar","Apr","May","Jun","Jul","Aug","Sep","Oct","Nov","Dec"};
	
char *a_homepage="www/index.html";
int a_last=0;
int a_first=0;
char *a_thumbnail_alt=NULL;
char *a_join_fields="urdm";
char *a_sort=NULL;
char *a_hostname=NULL;
char *a_mailname=NULL;
char *a_etc_passwd=NULL;
char *a_execmatch=NULL;
char *a_index=NULL;
char *a_index_caption=NULL;
char *a_index_tab_caption=NULL;
struct schain s_exclude_base, *s_exclude=&s_exclude_base;
struct schain s_match_base, *s_match=&s_match_base;
struct schain s_fexclude_base, *s_fexclude=&s_fexclude_base;
struct schain s_fmatch_base, *s_fmatch=&s_fmatch_base;
struct schain s_nexclude_base, *s_nexclude=&s_nexclude_base;
struct schain s_nmatch_base, *s_nmatch=&s_nmatch_base;
struct schain s_sort_base, *s_sort=&s_sort_base;
struct schain s_output_base, *s_output=&s_output_base;
struct schain s_caption_base, *s_caption=&s_caption_base;
struct schain s_pre_input_base, *s_pre_input=&s_pre_input_base;
struct schain s_thumbnail_base, *s_thumbnail=&s_thumbnail_base;
int f_order=0;
int f_ignorecase=0;
int f_substring=1;
int f_root=0;
int f_dir=0;
int f_user=0;
int f_no_stats=0;
int f_use_tables=0;
int f_use_2and3=0;
int f_mailto=0;
int f_scan_home_dir=0;
int f_scan_www_dir=0;
int f_blocks=0;
int f_prop_end=0;
int f_home_dir=0;
int f_include_no_home=0;
int f_mtime_all_files=0;
int f_pre=0;
int f_dont_check_home_page=0;

#define FLIP(y) (f_sort_rev? (y)*-1: (y))
int f_sort_rev=0;

char *wwwdir=NULL;
time_t tim;
struct tm *tm;
int stat_accounts=0;
int stat_home_pages=0;
int stat_bad=0;
int stat_selected=0;
int stat_pattern_pass=0;
int stat_pattern_fail=0;
int stat_exec_pass=0;
int stat_exec_fail=0;
int stat_time_fail=0;
int stat_total_age=0;
int stat_total_www_age=0;
int stat_total_files=0;
int stat_total_file_size=0;
int stat_total_www_files=0;
int stat_total_www_file_size=0;
int stat_files_included=0;
int stat_files_excluded=0;

FILE *findex=NULL;
char last_output_file[1024]="";

struct ent
{
	struct ent *next;
	char *name;
	char *realname;
	char *hostname;
	char *mailname;
	char *thumbnail;
	int f_found_home;
	char *dir;
	int uid;
	int files;
	int size;
	time_t dir_mtime;
	time_t www_mtime;
	int www_files;
	int www_size;
	int hid_numlogins;
	time_t hid_lastlogin;
	int hid_timeleft;
	int hid_totaltime;
	int hid_credit_used;
};

struct ent ent_base, *ent=&ent_base;

char *
stolower(char *ss)
{
	char *s=ss;
	for (;*s; s++) 
		*s=tolower(*s);
	return ss;
}

void *
xmalloc(int n)
{
	void *p;
	if (!(p=malloc(n)))
	{
		perror("malloc");
		exit(1);
	}
	return p;
}

char *
xstrdup(char *s)
{
	char *p;
	p=xmalloc(strlen(s)+1);
	strcpy(p, s);
	return p;
}

struct schain *
sadd(struct schain **sc, char *s)
{
	struct schain *p=*sc;
	p->str=s;
	p->next=xmalloc(sizeof (*p));
	p=p->next;
	p->next=NULL;
        p->str=NULL;
	*sc=p;
	return p;
}

float
con(float n, char *c)
{
	if (n>1073741824)
	{
		n/=1073741824.0;
		*c='G';
		return n;
	}
	if (n>1048576.0)
	{
		n/=1048567.0;
		*c='M';
		return n;
	}
	n/=1024.0;
	*c='k';
	return n;
}

char *
timetos(time_t *ti)
{
	static char buf[19];
	struct tm *tm;
	if (*ti)
	{
		tm = localtime(ti);
		sprintf(buf, "%02d:%02d %s %02d %04d", tm->tm_hour, tm->tm_min, months[tm->tm_mon], tm->tm_mday, tm->tm_year+1900);
	} else
		strcpy(buf, "n/a");
	return buf;
}

char *strrncpy(char *d, char *s, int n)
{

	char *p;
	for (p=d; *p && isspace(*p); p++);
	strncpy(p-strlen(s), s, strlen(s));
	return d;
}

char *
ttos (signed long id)
{
	static char buf[64];
	char *p=buf;
	signed long od=id;
	memset(buf, ' ', sizeof(buf)-1);
	if (od<0)
		return "        n/a";
	if (od<60)
	{
		return "         0<b>m</b>";
	}
	if (od > 86400)
	{
		char tmp[20];
		sprintf (tmp, "%ld<b>d</b>", id / (86400));
		sprintf(p, "%12s", tmp);
		p+=12;
	} else p+=5;
	id %= 86400;
	if (od > 3600)
	{
		sprintf (p, "%02ld<b>h</b>", id / 3600);
		p+=10;
	} else p+=3;
	id %= 3600;
	sprintf (p, "%02ld<b>m</b>", id / 60);
	p+=10;
	for (p=buf; *p==' ' || *p=='0'; p++)
		if (*p=='0')
			*p=' ';
	return buf;
}

int matches(char *file, struct schain *match)
{
	FILE *fp;
	char line[4096];
	if (!(fp=fopen(file, "r")))
		return -1;
	while (fgets(line, sizeof(line), fp))
	{
		struct schain *m;
		if (f_ignorecase)
			stolower(line);
		for (m=match; m && m->str; m=m->next)
		{
			char *p;
			if ((p=re_comp(m->str)))
			{
				perror(p);
				exit(1);
			}
			if (f_substring? !!strstr(line, m->str): re_exec(line))
			{
				fclose(fp);
				return 1;
			}
		}
	}
	fclose(fp);
	return 0;
}

int
sort_numlogins(struct ent **e1, struct ent **e2)
{
	int i=(*e1)->hid_numlogins - (*e2)->hid_numlogins;
	if (i<0) return FLIP(-1);
	return FLIP(i>0);
}

int
sort_lastlogin (struct ent **e1, struct ent **e2)
{
	int i=(*e1)->hid_lastlogin - (*e2)->hid_lastlogin;
	if (i<0) return FLIP(-1);
	return FLIP(i>0);
}

int
sort_timeleft (struct ent **e1, struct ent **e2)
{
	int i=(*e1)->hid_timeleft - (*e2)->hid_timeleft;
	if (i<0) return FLIP(-1);
	return FLIP(i>0);
}

int
sort_totaltime (struct ent **e1, struct ent **e2)
{
	int i=(*e1)->hid_totaltime - (*e2)->hid_totaltime;
	if (i<0) return FLIP(-1);
	return FLIP(i>0);
}

int
sort_credit (struct ent **e1, struct ent **e2)
{
	int i=(*e1)->hid_credit_used - (*e2)->hid_credit_used;
	if (i<0) return FLIP(-1);
	return FLIP(i>0);
}

int
sort_thumbnail(struct ent **e1, struct ent **e2)
{
	int i=((*e1)->thumbnail!=NULL) - ((*e2)->thumbnail!=NULL);
	if (i<0) return FLIP(-1);
	return FLIP(i>0);
}

int
sort_size(struct ent **e1, struct ent **e2)
{
	int i=(*e1)->size - (*e2)->size;
	if (i<0) return FLIP(-1);
	return FLIP(i>0);
}

int
sort_files(struct ent **e1, struct ent **e2)
{
	int i=(*e1)->files - (*e2)->files;
	if (i<0) return FLIP(-1);
	return FLIP(i>0);
}

int
sort_wwwsize(struct ent **e1, struct ent **e2)
{
	int i=(*e1)->www_size - (*e2)->www_size;
	if (i<0) return FLIP(-1);
	return FLIP(i>0);
}

int
sort_wwwfiles(struct ent **e1, struct ent **e2)
{
	int i=(*e1)->www_files - (*e2)->www_files;
	if (i<0) return FLIP(-1);
	return FLIP(i>0);
}

int
sort_uid(struct ent **e1, struct ent **e2)
{
	int i=(*e1)->uid - (*e2)->uid;
	if (i<0) return FLIP(-1);
	return FLIP(i>0);
}

int
sort_www_mtime(struct ent **e1, struct ent **e2)
{
	int i=(*e1)->www_mtime - (*e2)->www_mtime;
	if (i<0) return FLIP(-1);
	return FLIP(i>0);
}

int
sort_www_days(struct ent **e1, struct ent **e2)
{
	int i=(*e2)->www_mtime - (*e1)->www_mtime;
	if (i<0) return FLIP(-1);
	return FLIP(i>0);
}

int
sort_dir_mtime(struct ent **e1, struct ent **e2)
{
	int i=(*e1)->dir_mtime - (*e2)->dir_mtime;
	if (i<0) return FLIP(-1);
	return FLIP(i>0);
}

int
sort_dir_days(struct ent **e1, struct ent **e2)
{
	int i=(*e2)->dir_mtime - (*e1)->dir_mtime;
	if (i<0) return FLIP(-1);
	return FLIP(i>0);
}

int
sort_name(struct ent **e1, struct ent **e2)
{
	return FLIP(strcmp((*e1)->name, (*e2)->name));
}

int
sort_realname(struct ent **e1, struct ent **e2)
{
	char *s1=(*e1)->realname;
	char *s2=(*e2)->realname;
	char *p;
	for (p=s1; *p; p++)
		if (isspace(*p)) s1=p+1;
	for (p=s2; *p; p++)
		if (isspace(*p)) s2=p+1;
	return FLIP(strcasecmp(s1, s2));
}

struct sortby
{
	int (*func)(void *, void *);
	char *header;
	int direction;
	char *desc;
	int max;
} sortby[] = {
	{},{}, /* a,b */
	(struct sortby){sort_credit, "Credit", RIGHT, "credit", 6}, /* c */
	(struct sortby){sort_dir_days, "Age", RIGHT, "age in days", 4}, /* d */
	(struct sortby){sort_timeleft, "Time left", RIGHT, "timeleft", 11}, /* e */
	(struct sortby){sort_files, "Files", RIGHT, "number of files", 5}, /* f */
	{},{}, /* g,h */
	(struct sortby){sort_uid, "UID", RIGHT, "user ID", 5}, /* i */
	{},{}, /* j,k */
	(struct sortby){sort_lastlogin, "Lastlogin", RIGHT, "lastlogin", 17}, /* l */
	(struct sortby){sort_dir_mtime, "Last modified", RIGHT, "last modification", 17}, /* m */
	(struct sortby){sort_numlogins, "Logins", RIGHT, "number of logins", 6}, /* n */
	(struct sortby){sort_totaltime, "Time Used", RIGHT, "time online", 11}, /* o */
	{}, /* p */
	(struct sortby){sort_wwwfiles, "W3Files", RIGHT, "number of www files", 7}, /* q */
	(struct sortby){sort_realname, "Full name", LEFT, "full name", 9}, /* r */
	(struct sortby){sort_size, "Usage", RIGHT, "disk usage", 6}, /* s */
	(struct sortby){sort_thumbnail, "Thumbnail", CENTER, "thumbnail", 20}, /* t */
	(struct sortby){sort_name, "Username", LEFT, "user name", 8}, /* u */
	{},
	(struct sortby){sort_www_mtime, "W3 Last update", RIGHT, "last www modification", 17}, /* w */
	(struct sortby){sort_www_days, "W3Age", RIGHT, "www age in days", 5}, /* x */
	{},
	(struct sortby){sort_wwwsize, "W3Size", RIGHT, "www disk usage", 6} /* z */
}; /* yow */

void
printent(struct ent *e)
{
	char c;
	float f=0.0;
	char *p=NULL;
	int dir_days, www_days;
	char dir_ds[32];
	char www_ds[32];
	if (e->dir_mtime)
	{
		if ((dir_days = (tim-e->dir_mtime)/84600)<0)
			dir_days=0;
		sprintf(dir_ds, "%d", dir_days);
	} else
	{
		dir_days=-1;
		strcpy(dir_ds, "n/a");
	}
	if (e->www_mtime)
	{
		if ((www_days = (tim-e->www_mtime)/84600)<0)
			www_days=0;
		sprintf(www_ds, "%d", www_days);
	} else
	{
		strcpy(www_ds, "n/a");
		www_days=-1;
	}
	if (f_user)
	{
		printf("%s:%d:%s:%s:%s:%s:%s:%d:%lu:%d:%d:%lu:%d:%d:%d:%lu:%d:%d:%d\n", e->name, e->uid, e->realname, e->hostname, e->mailname, e->dir, e->thumbnail? e->thumbnail: "", e->f_found_home, e->dir_mtime, e->files, e->size, e->www_mtime, e->www_files, e->www_size, e->hid_numlogins, e->hid_lastlogin, e->hid_timeleft, e->hid_totaltime, e->hid_credit_used);
		return;
	}
	if (f_use_tables)
	{
		printf("<tr>");
		for (p=a_join_fields; *p; printf("</td>"), p++)
		{
			struct sortby *s;
			s=&sortby[*p-'a'];
			if (!s->func)
			{
				fprintf(stderr, "invalid join field (-j) flag '%c'\n", *p);
				exit(1);
			}
			printf("<td align=%s>", align[s->direction]);
			switch(*p)
			{
			case 'c':
				if (e->hid_credit_used<0)
					printf("n/a");
				else
					printf("%d", e->hid_credit_used);
				break;
			case 'd':
				printf("%s", dir_ds);
				break;
			case 'e':
				printf("%s", ttos(e->hid_timeleft));
				break;
			case 'f':
				printf("%d", e->files);
				break;
			case 'i':
				printf("%d", e->uid);
				break;
			case 'l':
				printf("%s", timetos(&e->hid_lastlogin));
				break;
			case 'm':
				printf("%s", timetos(&e->dir_mtime));
				break;
			case 'n':
				if (e->hid_numlogins<0)
					printf("n/a");
				else
					printf("%d", e->hid_numlogins);
				break;
			case 'o':
				printf("%s", ttos(e->hid_totaltime));
				break;
			case 'q':
				printf("%d", e->www_files);
				break;
			case 'r':
				if (e->f_found_home)
					printf("<a href=\"http://%s/~%s\">%s</a>", e->hostname, e->name, e->realname);
				else
					printf("%s", e->realname);
				break;
			case 's':
				f=con(e->size, &c);
				printf("%.1f<b>%c</b>", f, c);
				break;
			case 't':
				if (!e->thumbnail && !a_thumbnail_alt)
					break;
				printf("<a href=\"http://%s/~%s\">", e->hostname, e->name);
				if (e->thumbnail)
					printf("<IMG SRC=\"http://%s/~%s/%s\" ALT=Home>", e->hostname, e->name, e->thumbnail);
				else
					printf("%s", a_thumbnail_alt);
				printf("</a>");
				break;
			case 'u':
				if (f_mailto)
					printf("<a href=\"mailto:%s@%s\">%s</a>", e->name, e->mailname, e->name);
				else
					printf("%s", e->name);
				break;
			case 'w':
				printf("%s", timetos(&e->www_mtime));
				break;
			case 'x':
				printf("%s", www_ds);
				break;
			case 'z':
				f=con(e->www_size, &c);
				printf("%.1f<b>%c</b>", f, c);
				break;
			default:
				fprintf(stderr, "invalid join field (-j) flag '%c'\n", *p);
				exit(1);
			}
		}
		puts("</tr>");
	} else {
		int cnt=0;
		for (p=a_join_fields; *p; p++)
		{
			struct sortby *s;
			if (!(s=&sortby[*p-'a']))
			{
				fprintf(stderr, "invalid join field (-j) flag '%c'\n", *p);
				exit(1);
			}
			if (cnt++)
				putchar(' ');
			switch(*p)
			{
			case 'c':
				if (e->hid_credit_used<0)
					printf("   n/a");
				else
					printf("%6d", e->hid_credit_used);
				break;
			case 'd':
				printf("%4s", dir_ds);
				break;
			case 'e':
				printf("%s", ttos(e->hid_timeleft));
				break;
			case 'f':
				printf("%5d", e->files);
				break;
			case 'i':
				printf("%5d", e->uid);
				break;
			case 'l':
				printf("%17s", timetos(&e->hid_lastlogin));
				break;
			case 'm':
				printf("%17s", timetos(&e->dir_mtime));
				break;
			case 'n':
				if (e->hid_numlogins<0)
					printf("   n/a");
				else
					printf("%6d", e->hid_numlogins);
				break;
			case 'o':
				printf("%s", ttos(e->hid_totaltime));
				break;
			case 'q':
				printf("%7d", e->www_files);
				break;
			case 's':
				f=con(e->size, &c);
				printf("%5.1f<b>%c</b>", f, c);
				break;
			case 't':
				printf("<a href=\"http://%s/~%s\">", e->hostname, e->name);
				if (e->thumbnail)
					printf("<IMG SRC=\"http://%s/~%s/%s\" ALT=Home>", e->hostname, e->name, e->thumbnail);
				else if (a_thumbnail_alt)
						printf("%s", a_thumbnail_alt);
					else
						printf("Home");
				printf("%*s</a>", s->max-(a_thumbnail_alt? 0: 4), "");
				break;
			case 'u':
				if (f_mailto)
					printf("<a href=\"mailto:%s@%s\">%-*s</a>", e->name, e->mailname, s->max, e->name);
				else
					printf("%-*s", s->max, e->name);
				break;
			case 'r':
				if (e->f_found_home)
					printf("<a href=\"http://%s/~%s\">%-*s</a>", e->hostname, e->name, s->max, e->realname);
				else
					printf("%-*s", s->max, e->realname);
				break;
			case 'w':
				printf("%17s", timetos(&e->www_mtime));
				break;
			case 'x':
				printf("%5s", www_ds);
				break;
			case 'z':
				f=con(e->www_size, &c);
				printf("%5.1f<b>%c</b>", f, c);
				break;
			default:
				fprintf(stderr, "invalid join field (-j) flag '%c'\n", *p);
				exit(1);
			}
		}
		putchar('\n');
	}
}



int
dir_stats(char *path, int *numfiles, int *size, time_t *mtime)
{
	DIR *dir;
	struct dirent *d;
	struct stat st;
	char *p;
	int path_len;
	int high=0;
	if (!strcmp(path, "/")) /* eeeeew */
		return -1;
	if (!(dir=opendir(path)))
		return -1;
	path_len=strlen(path);
	p=xmalloc(path_len+2);
	sprintf(p, "%s/", path);
	while ((d=readdir(dir)))
	{
		int n;
		int f_included=0;
		struct schain *m;
		
		if (!d->d_name || !d->d_name[0] || !strcmp(d->d_name, ".") || !strcmp(d->d_name, ".."))
			continue;
		n=strlen(d->d_name);
		if (n>high)
		{
			high=n;
			p=realloc(p, path_len+2+high);
			if (!p)
			{
				perror("out of memory");
				exit(1);
			}
		}
		strcpy(p+path_len+1, d->d_name);
		for (m=s_fexclude; m && m->str; m=m->next)
		{
			char *s;
			if ((s=re_comp(m->str)))
			{
				perror(s);
				exit(1);
			}
			if (re_exec(p))
			{
				stat_files_excluded++;
				goto ewhile;
			}
		}
		for (m=s_fmatch; m && m->str; m=m->next)
		{
			char *s;
			if ((s=re_comp(m->str)))
			{
				perror(s);
				exit(1);
			}
			if (!re_exec(p))
				continue;
			else
			{
				stat_files_included++;
				f_included=1;
				break;
			}
		}
		if (s_fmatch->str && !f_included)
			continue;
		if (lstat(p, &st)==-1)
			continue;
		if (f_blocks)
			*size+=st.st_blocks * st.st_blksize/2;
		(*numfiles)++;
		if (S_ISREG(st.st_mode))
		{
			if (f_mtime_all_files && st.st_mtime>*mtime)
				*mtime=st.st_mtime;
			if (!f_blocks)
				*size+=st.st_size;
			continue;
		}
		if (S_ISDIR(st.st_mode))
		{
			if (f_dir && st.st_mtime>*mtime)
				*mtime=st.st_mtime;
			dir_stats(p, numfiles, size, mtime);
		}
ewhile:
	}
	free(p);
	closedir(dir);
	return 0;
}

void
open_son()
{
static int cnt;
	if (s_output && s_output->str)
	{
		char *p;
		char *s, *ss;
		if (cnt || (p=strstr(s_output->str, "%d")))
		{
		        s=ss=xmalloc(strlen(s_output->str)+25);
			sprintf(s, s_output->str, cnt++);
		} else {
			s=s_output->str;
			ss=NULL;
		}
		if (!(freopen(s, "w", stdout)))
		{
			perror(s);
			exit(1);
		}
		strncpy(last_output_file, s, sizeof(last_output_file));
		if (ss)
			free(ss);
		if ((s_output->next && s_output->next->str) || !cnt)
			s_output=s_output->next;
	}
}


void
properganda()
{
	if (f_use_tables)
		printf("<center><table border=4><tr><td><a href=http://suburbia.net/homealone/homealone.html><b>HomeAlone</b> <em>%s</em></td><td>(c) 1995 Julian Assange (<a href=mailto:proff@suburbia.net>proff@suburbia.net</a>)</td><td>Serial #%s</td></tr></table></centre>\n", VERSION, SERIAL);
	else
		printf("<center>This index was generated by <a href=http://suburbia.net/homealone/homealone.html><b>HomeAlone</b></a> <em>%s</em> (%s)<br><address>(c) 1995 Julian Assange (<a href=mailto:proff@suburbia.net>proff@suburbia.net</a>)</address></center>\n", VERSION, SERIAL);
}

void
header(char *desc)
{
	open_son();
	if (!f_user)
	{
		char *cap=NULL;
		char *p;
		int cnt;
		if (!s_caption->str)
		{
			cap=xmalloc(128);
			sprintf(cap, "Users sorted by <em>%s</em> on <date>%s</date>", desc, ctime(&tim));
		}
		else
		{
			cap=xmalloc(strlen(s_caption->str)+strlen(desc)+128); /* 128 incase some bozo wants to do %32s */
			sprintf(cap, s_caption->str, desc);
		}
		if (findex)
		{
			char *icap;
			if (a_index_tab_caption && f_use_tables)
				icap=a_index_tab_caption;
			else
				icap=a_index_caption;
			if (icap)
			{
				char buf[strlen(icap)+128];
				sprintf(buf, icap, last_output_file, desc);
				fprintf(findex, "<li>%s\n", buf);
			} else
				fprintf(findex, "<li>%s\n", cap);
		}
		if (!f_prop_end)
		{
			if (!f_use_tables)
				puts("<hr>");
			properganda();
			if (f_use_tables)
				puts("<p>");
			else
				puts("<hr>");
		}
		if (f_use_tables)
			printf("<center><table border=2><caption align=top>%s</caption><tr>\n", cap);
		else
		{
			printf("<center>%s</center>\n", cap);
			if (!f_prop_end)
				puts("<p>");
			printf("<pre><em>");
		}
		for (cnt=0, p=a_join_fields; *p; p++)
		{
			struct sortby *s;
			if (!(s=&sortby[*p-'a']))
			{
				fprintf(stderr, "invalid join field (-j) flag '%c'\n", *p);
				exit(1);
			}
			if (f_use_tables)
				printf("<th align=%s>%s</th>", align[s->direction], s->header);
			else
			{
				if (cnt++)
					putchar(' ');
				if (s->direction==RIGHT)
					printf("%*s", s->max, s->header);
				else
					printf("%-*s", s->max, s->header);
			}
		}
		if (f_use_tables)
			puts("</tr>");
		else puts("</em>");
		free(cap);
	}
}

void
footer()
{
	if (f_user)
		return;
	if (f_use_tables)
	{
		puts("</table>");
		if (f_prop_end)
		{
			puts("<p>");
			properganda();
		}
		puts("</center>");
	} else {
		puts("</pre>");
		if (f_prop_end)
		{
			puts("<hr>");
			properganda();
		}
	}
	puts("<hr>");
}

int
hload(char *p1, char *p2, char *name, void *i, int arg, int f_unsigned)
{
	char buf[32], *p=buf;
	int fd;
	strcpy(p2, name);
	if ((fd=open(p1, O_RDONLY))==-1)
		return 0;
	if (read(fd, buf, 32)<1)
	{
		close(fd);
		return 0;
	}
	close(fd);
	for (; arg; arg--)
	{
		for (; *p && !isspace(*p); p++);
		for (; *p && isspace(*p); p++);
	}
	if (f_unsigned)
	{
		if (sscanf(p, "%lu", (unsigned long *)i)!=1)
			return 0;
	} else 
		if (sscanf(p, "%d", (int *)i)!=1)
			return 0;
	return 1;
}

void
stats()
{
	if (!f_user)
	{
		if (!f_no_stats)
		{
			char c1, c2;
			float f1, f2;
			f1=con(stat_total_file_size, &c1);
			f2=con(stat_total_files? (float)stat_total_file_size/stat_total_files: 0 , &c2);
			open_son();
			if (!f_use_tables)
				printf("<h3>Home page statistics</h3>\n<pre>\n\
Accounts             %8d\n\
Home pages           %8d\n\
Bad home pages       %8d\n\
Matched pattern      %8d\n\
Failed pattern       %8d\n\
Matched exec pattern %8d\n\
Failed exec pattern  %8d\n\
Files scanned        %8d\n\
Files included       %8d\n\
Files excluded       %8d\n\
Space consumed      %8.1f%c\n\
Avg. file size      %8.1f%c\n\
Failed time ranges   %8d\n\
Avg page age (days)  %8.1f\n\
Scan time (seconds)  %8d\n\
Accounts/second      %8.1f\n\
-----------------------------\n\
Total selected       %8d\n\
-----------------------------\n\
</pre>",
				stat_accounts, stat_home_pages, stat_bad, stat_pattern_pass, stat_pattern_fail,
					stat_exec_pass, stat_exec_fail, stat_total_files, stat_files_included, stat_files_excluded,
					f1, c1, f2, c2, stat_time_fail, stat_selected? (float)stat_total_age/(float)stat_selected: 0, (int)(time(NULL)-tim), (float)stat_accounts/(float)(time(NULL)-tim+1), stat_selected);
			else
				printf("<center><table border=3 cellspacing=5><tr><th align=left></th><th align=left>quantity</th></tr>\n\
<tr><th align=left>Accounts</th><td align=right>%d</td></tr>\n\
<tr><th align=left>Home pages</th><td align=right>%d</td></tr>\n\
<tr><th align=left>Bad home pages</th><td align=right>%d</td></tr>\n\
<tr><th align=left>Matched pattern</th><td align=right>%d</td></tr>\n\
<tr><th align=left>Failed pattern</th><td align=right>%d</td></tr>\n\
<tr><th align=left>Matched exec pattern</th><td align=right>%d</td></tr>\n\
<tr><th align=left>Failed exec pattern</th><td align=right>%d</td></tr>\n\
<tr><th align=left>Files scanned</th><td align=right>%d</td></tr>\n\
<tr><th align=left>Files included</th><td align=right>%d</td></tr>\n\
<tr><th align=left>Files excluded</th><td align=right>%d</td></tr>\n\
<tr><th align=left>Space consumed</th><td align=right>%.1f<b>%c</b></td></tr>\n\
<tr><th align=left>Avg. file size</th><td align=right>%.1f<b>%c</b></td></tr>\n\
<tr><th align=left>Failed time ranges</th><td align=right>%d</td></tr>\n\
<tr><th align=left>Avg. page age (days)</th><td align=right>%.1f</td></tr>\n\
<tr><th align=left>Scan time (seconds)</th><td align=right>%d</td></tr>\n\
<tr><th align=left>Accounts/second</th><td align=right>%.1f</td></tr>\n\
<tr><th align=left><font size=+2>Total selected</font></th><td align=right><font size=+2>%d</font></td></tr></font>\n\
</table></center>",
				stat_accounts, stat_home_pages, stat_bad, stat_pattern_pass, stat_pattern_fail,
					stat_exec_pass, stat_exec_fail, stat_total_files, stat_files_included, stat_files_excluded, f1, c1, f2, c2, stat_time_fail, stat_selected? (float)stat_total_age/(float)stat_selected: 0, (int)(time(NULL)-tim), (float)stat_accounts/(float)(time(NULL)-tim+1), stat_selected);
		}
	}
}



void
usage(char *s)
{
	fprintf(stderr, "usage: %s [-3AbBdDEiImnNrRuWx] [-acefGgGhHjlLMopPstTvVw arg]
	-a address
	-c caption
	-e command
	-f pattern
	-F pattern
	-g pattern
	-G pattern
	-h path
	-H directory
	-j [cdefilmnoqrstuwxz]
	-l days
	-L days
	-M address
	-o file
	-p file
	-P file
	-s [cCdDeEfFiIlLmMnNoOqQrRsStTuUwWxXzZ]
	-t thumbnail
	-T URL
	-v pattern
	-V pattern
	-w width

	See the manual page for full information.
	(c) 1995 Julian Assange (proff@suburbia.net)\n", s);
	exit(1);
}

int
main(int argc, char **argv)
{
	char s1[4096+3]="in=", *s=s1+3; /* space for putenv("in="); */
	char s2[4096]; /* enough to hold a file path */
	struct stat st;
	struct passwd *pw;
	FILE *pwin=NULL, *fpre=NULL;
	int c;
	struct schain *mat1, *mat2;
	char *p;
	char *p_hidleho=NULL, *p2_hidleho=NULL;
	int pre_cnt=0;
	while ((c=getopt(argc, argv, "h:a:v:V:g:G:e:rixRds:p:unl:L:t:T:3BmM:Wbo:c:EDIP:h:f:F:AH:j:Nk:K:U:")) != -1)
		switch(c)
		{
		case 'h':
			a_homepage=optarg;
			break;
		case 'a':
			a_hostname=optarg;
			break;
		case 'v':
			sadd(&s_exclude, optarg);
			break;
		case 'V':
			sadd(&s_fexclude, optarg);
			break;
		case 'g':
			sadd(&s_match, optarg);
			break;
		case 'G':
			sadd(&s_fmatch, optarg);
			break;
		case 'F':
			sadd(&s_nexclude, optarg);
			break;
		case 'f':
			sadd(&s_nmatch, optarg);
			break;
		case 'e':
			a_execmatch=optarg;
			break;
		case 'r':
			f_order=1;
			break;
		case 'i':
			f_ignorecase=1;
			break;
		case 'x':
			f_substring=0;
			break;
		case 'R':
			f_root=1;
			break;
		case 'd':
			f_dir=1;
			break;
		case 's':
			sadd(&s_sort, optarg);
			break;
		case 'p':
			a_etc_passwd=optarg;
			break;
		case 'u':
			f_user=1;
			break;
		case 'n':
			f_no_stats=1;
			break;
		case 'U':
			a_index_tab_caption=optarg;
			break;
		case 'k':
			a_index_caption=optarg;
			break;
		case 'K':
			a_index=optarg;
			break;
		case 'l':
			a_first=atoi(optarg);
			break;
		case 'L':
			a_last=atoi(optarg);
			break;
		case 'B':
			f_use_2and3=1;
			break;
		case '3':
			f_use_tables=1;
			break;
		case 'M':
			a_mailname=optarg;
		case 'm':
			f_mailto=1;
			break;
		case 'D':
			f_scan_home_dir=1;
			break;
		case 'W':
			f_scan_www_dir=1;
			break;
		case 'o':
			sadd(&s_output, optarg);
			break;
		case 'b':
			f_blocks=1;
			break;
		case 'c':
			sadd(&s_caption, optarg);
			break;
		case 'E':
			f_prop_end=1;
			break;
		case 'I':
			f_include_no_home=1;
			break;
		case 'P':
			sadd(&s_pre_input, optarg);
			f_pre=1;
			break;
		case 'A':
			f_mtime_all_files=1;
			break;
		case 'H':
			p_hidleho=xmalloc(strlen(optarg)+1+32+1);
			sprintf(p_hidleho, "%s/", optarg);
			p2_hidleho=p_hidleho+strlen(optarg)+1;
			break;
		case 'j':
			a_join_fields=stolower(optarg);
			break;
		case 'N':
			f_dont_check_home_page=1;
			break;
		case 't':
			sadd(&s_thumbnail, optarg);
			break;
		case 'T':
			a_thumbnail_alt=optarg;
			break;
		default:
			fprintf(stderr, "%s: unknown option '%c'\n", argv[0], c);
			usage(argv[0]);
		}
	s_caption=&s_caption_base;
	s_output=&s_output_base;
	s_sort=&s_sort_base;
	s_exclude=&s_exclude_base;
	s_match=&s_match_base;
	s_fexclude=&s_fexclude_base;
	s_fmatch=&s_fmatch_base;
	s_nexclude=&s_nexclude_base;
	s_nmatch=&s_nmatch_base;
	s_pre_input=&s_pre_input_base;
	s_thumbnail=&s_thumbnail_base;
	if (!a_hostname)
	{
		char buf[256];
		if (gethostname(buf, 256)==-1)
		{
			perror("couldn't obtain host name");
			exit(1);
		}
		a_hostname=xstrdup(buf);
		if (!strchr(a_hostname, '.'))
		{
			struct hostent *h;
			if ((h=gethostbyname(a_hostname)))
				a_hostname=(char *)h->h_name;
			if (!strchr(a_hostname, '.'))
				fprintf(stderr, "%s warning: gethostname() address '%s' is not a FQDN!\n", argv[0], a_hostname);
		}
	}
	if (!a_mailname)
		a_mailname=a_hostname;
	wwwdir=xstrdup(a_homepage);
	if (!(p=strrchr(wwwdir, '/')))
	{
		free(wwwdir);
		wwwdir=NULL;
	}
	*p='\0';
	mat1=f_order? (s_match_base.str? &s_match_base: NULL): (s_exclude_base.str? &s_exclude_base: NULL);
	mat2=f_order? (s_exclude_base.str? &s_exclude_base: NULL): (s_match_base.str? &s_match_base: NULL);
	if (f_ignorecase)
	{
		struct schain *p;
		for (p=mat1; p && p->str; p=p->next)
			stolower(p->str);
		for (p=mat2; p && p->str; p=p->next)
			stolower(p->str);
	}
	tim=time(NULL);
	if (a_etc_passwd && !f_pre)
	{
		if (!(pwin=fopen(a_etc_passwd, "r")))
		{
			perror(a_etc_passwd);
			exit(1);
		}
	}
	if (!s_sort->str)
	{
		sortby['r'-'a'].max=40; /* r */
		sortby['u'-'a'].max=8; /* u */
		header("account order");
	}
	for (; (!f_pre && !pre_cnt) || (s_pre_input && s_pre_input->str); s_pre_input=s_pre_input->next, pre_cnt++, fclose(fpre))
	{
		if (f_pre && !(fpre=fopen(s_pre_input->str, "r")))
		{
			perror(s_pre_input->str);
			exit(1);
		}
		for (;; stat_accounts++)
		{
			int mat;
			int t;
			char *p;
			int f_found_home=0;
			int f_included=0;
			struct schain *m;
			char name[32], realname[256], hostname[256]="", mailname[256]="", thumbnail[256]="", dir[4097];
			bzero(ent, sizeof(*ent));
			ent->hostname=a_hostname;
			ent->mailname=a_mailname;
			ent->hid_numlogins=ent->hid_timeleft=ent->hid_totaltime=ent->hid_credit_used=-1;
			if (f_pre)
			{
				char line[4096];
				struct passwd pwd;
				int i=0;

				pw=&pwd;
				if (!fgets(line, sizeof(line), fpre))
					break;
				*realname=*name=*dir='\0';
				sscanf(line, "%31[^:]:%d:%255[^:]:%255[^:]:%255[^:]:%4096[^:]:%255[^:]:%d:%lu:%d:%d:%lu:%d:%d:%d:%lu:%d:%d:%d", name, &i, realname, hostname, mailname, dir, thumbnail, &f_found_home, &ent->dir_mtime, &ent->files, &ent->size, &ent->www_mtime, &ent->www_files, &ent->www_size, &ent->hid_numlogins, &ent->hid_lastlogin, &ent->hid_timeleft, &ent->hid_totaltime, &ent->hid_credit_used);
				pw->pw_uid=i;
				pw->pw_name=name;
				pw->pw_gecos=realname;
				pw->pw_dir=dir;
				if (*hostname)
					ent->hostname=xstrdup(hostname);
				if (*mailname)
					ent->mailname=xstrdup(hostname);
				if (*thumbnail)
					ent->thumbnail=xstrdup(hostname);
			} else {
				if (!(pw=(a_etc_passwd? fgetpwent(pwin): getpwent())))
					break;
			}
			if (!pw->pw_name || !*pw->pw_name)
				continue;
			if (!pw->pw_dir || !*pw->pw_dir)
				pw->pw_dir="/";
			if (!pw->pw_gecos || !*pw->pw_gecos)
				pw->pw_gecos=pw->pw_name;
			else
				for (p=pw->pw_gecos; *p; p++)
					if (*p==',')
					{
						*p='\0';
						break;
					}
			for (m=s_nexclude; m && m->str; m=m->next)
			{
				char *s;
				if ((s=re_comp(m->str)))
				{
					perror(s);
					exit(1);
				}
				if (re_exec(pw->pw_name) || re_exec(pw->pw_gecos))
				{
					continue;
					goto eloop;
				}
			}
			for (m=s_nmatch; m && m->str; m=m->next)
			{
				char *s;
				if ((s=re_comp(m->str)))
				{
					perror(s);
					exit(1);
				}
				if (!re_exec(pw->pw_name) && !re_exec(pw->pw_gecos))
					continue;
				else
				{
					f_included=1;
					break;
				}
			}
			if (!f_dont_check_home_page || f_scan_www_dir || f_dir)
				sprintf(s, "%s/%s", pw->pw_dir, a_homepage);
			if (!f_pre && (!f_dont_check_home_page || f_scan_www_dir || f_dir))
			{
				if (stat(s, &st) == -1)
				{
					if (!f_include_no_home)
						continue;
				} else {
					stat_home_pages++;
					f_found_home=1;
					if (!f_root && !(st.st_mode & 4)) /* needs to be world readable */
					{
						stat_bad++;
						if (f_include_no_home)
							f_found_home=0;
						else
							continue;
					}
					ent->www_mtime=st.st_mtime;
				}
			}
			if (!f_found_home && !f_include_no_home)
				continue;
			if (!f_pre && (!f_root || f_dir || f_scan_home_dir))
			{
				if ((stat(pw->pw_dir, &st)==-1) || (!(st.st_mode & 1) && !f_root))
				{
					if (f_found_home)
					{
						stat_bad++;
						if (f_include_no_home)
							f_found_home=0;
						else
							continue;
					}
				} else
					if (f_dir)
						ent->dir_mtime=st.st_mtime;
			}
			if (!f_pre && !f_dont_check_home_page)
			{
				sprintf(s2, "%s/%s", pw->pw_dir, wwwdir);
				if (stat(s2, &st) == -1)
				{
					if (f_include_no_home)
						f_found_home=0;
					else
						continue;
				} else
				{
					int f_bad=0;
					if (!f_root && !(st.st_mode & 1)) /* needs to be world enterable */
					{
						stat_bad++;
						f_bad=1;
						if (f_include_no_home)
							f_found_home=0;
						else
							continue;
					}
					if (f_dir && st.st_mtime > ent->www_mtime)
						ent->www_mtime=st.st_mtime;
					if (!f_bad && s_thumbnail->str)
					{
						struct schain *sc;
						for (sc=s_thumbnail; sc && sc->str; sc=sc->next)
						{
							sprintf(s, "%s/%s/%s", pw->pw_dir, wwwdir, sc->str);
							if (!stat(s, &st) && (f_root || (st.st_mode&4)))
							{
								ent->thumbnail=sc->str;
								break;
							}
						}
					}
				}
			}
			if (a_last && ent->www_mtime && tim-ent->www_mtime<a_last*84600)
			{
				stat_time_fail++;
				if (f_include_no_home)
					f_found_home=0;
				else
					continue;
			}
			if (a_first && ent->www_mtime && tim-ent->www_mtime>a_first*84600)
			{
				stat_time_fail++;
				if (f_include_no_home)
					f_found_home=0;
				else
					continue;
			}
			if (f_found_home && mat1 && (((mat=matches(s, mat1)) == -1) || (mat ^ f_order)))
			{
				stat_pattern_fail++;
				if (f_include_no_home)
					f_found_home=0;
				else
					continue;
			}
			if (f_found_home && mat2 && (((mat=matches(s, mat2)) == -1) || (!mat ^ f_order)))
			{
				stat_pattern_fail++;
				if (f_include_no_home)
					f_found_home=0;
				else
					continue;
			}
			if (f_found_home && (mat1 || mat2))
				stat_pattern_pass++;
			if (f_found_home && a_execmatch)
			{
				putenv(s1);
				if (system(a_execmatch))
				{
					stat_exec_fail++;
					if (!f_include_no_home)
						continue;
					f_found_home=0;
				} else
					stat_exec_pass++;
			}
			if (p_hidleho)
			{
				struct stat sb;
				char *p3=p2_hidleho+strlen(pw->pw_name)+1;
				sprintf(p2_hidleho, "%s", pw->pw_name);
				if (!stat(p_hidleho, &sb))
				{
					*(p3-1)='/';
					hload(p_hidleho, p3, "numlogins", &ent->hid_numlogins, 0, 0);
					hload(p_hidleho, p3, "lastlogin", &ent->hid_lastlogin, 2, 1);
					hload(p_hidleho, p3, "timeleft", &ent->hid_timeleft, 0, 0);
					hload(p_hidleho, p3, "totaltime", &ent->hid_totaltime, 0, 0);
					hload(p_hidleho, p3, "credit", &ent->hid_credit_used, 1, 0);
				}
			}
			if (f_pre)
			{
				stat_total_files+=ent->files;
				stat_total_file_size+=ent->size;
				stat_total_www_files+=ent->files;
				stat_total_www_file_size+=ent->size;
			} else
			{
				time_t mtime=0;
				if (f_scan_home_dir && (dir_stats(pw->pw_dir, &(ent->files), &(ent->size), &mtime)!=-1))
				{
					stat_total_files+=ent->files;
					stat_total_file_size+=ent->size;
				}
				if (f_mtime_all_files && mtime>ent->dir_mtime)
					ent->dir_mtime=mtime;
				mtime=0;
				if (f_scan_www_dir && (dir_stats(s2, &(ent->www_files), &(ent->www_size), &mtime)!=-1))
				{
					stat_total_www_files+=ent->www_files;
					stat_total_www_file_size+=ent->www_size;
				}
				if (f_mtime_all_files && mtime>ent->www_mtime)
					ent->www_mtime=mtime;
			}
			ent->f_found_home=f_found_home;
			ent->dir=pw->pw_dir;
			if (ent->dir_mtime > tim)
				ent->dir_mtime=tim;
			if (ent->www_mtime > tim)
				ent->www_mtime=tim;
			stat_total_age+=(tim - ent->dir_mtime)/86400;
			stat_total_www_age+=(tim - ent->dir_mtime)/86400;
			ent->uid=pw->pw_uid;
			ent->name=s_sort_base.str? xstrdup(pw->pw_name): pw->pw_name;
			ent->realname=s_sort_base.str? xstrdup(pw->pw_gecos): pw->pw_gecos;
			if (s_sort_base.str)
			{
				t=strlen(ent->name);
				if (t>sortby['u'-'a'].max)
					sortby['u'-'a'].max=t;
				t=strlen(ent->realname);
				if (t>sortby['r'-'a'].max)
					sortby['r'-'a'].max=t;
				ent->next=xmalloc(sizeof(*ent));
				ent=ent->next;
			} else
				printent(ent);
				stat_selected++;
	eloop:
		}
	}
	bzero(ent, sizeof(*ent));
	if (s_sort_base.str)
	{
		char *p;
		struct schain *m;
		struct ent *e, **entp, **entb;
		if (stat_selected<1)
			exit(0);
		entp=entb=xmalloc(sizeof(*entp) *(stat_selected+1));
		for (e=&ent_base; e && e->name; e=e->next)
			*(entp++)=e;
		*entp=NULL;
		if (a_index)
		{
			if (!(findex=fopen(a_index, "w")))
			{
				perror(a_index);
				exit(1);

			}
			fprintf(findex, "<ul>\n");
		}
		for (m=&s_sort_base; m && m->str; m=m->next)
		{
			struct sortby *f=NULL;
			int f_second_loop=0;
			if (f_use_2and3)
				f_use_tables=0;
fakeloop:
			for (p=m->str; *p; p++);
			for (; p--!=m->str;)
			{
				char c=tolower(*p);
				f=&sortby[c-'a'];
				if (!f->func)
				{
					fprintf(stderr, "invalid sortby flag '%c'\n", *p);
					exit(1);
				}
				f_sort_rev=isupper(*p);
				if (stat_selected>1)
					qsort(entb, stat_selected, sizeof (*entb), f->func);
			}
			header(f->desc);
			for (entp=entb; *entp; entp++)
				printent(*entp);
			footer();
			if (f_use_2and3 && !f_second_loop)
			{
				f_use_tables=f_second_loop=1;
				goto fakeloop;
			}
		}
	} else 
		footer();
	if (f_use_2and3)
	{
		f_use_tables=0;
		stats();
		f_use_tables=1;
	}
	if (findex)
	{
		fprintf(findex, "</ul>\n");
		fclose(findex);
	}
	stats();
	exit(0);
}
